--[[
UDodge v1.0
by Brett "Megiddo" Smith

This script will create a "pylon" that shoots full-sized combine balls at you.

Requirements:
This script requires ULib version 1 or higher. You can download ULib from ulyssesmod.net

Usage:
Just use the command "makepylon" in console, a pylon will appear in front of you and start shooting!

For more advanced users, the full command is makepylon [area_size] [update_rate] [timeout] [ball_speed] [charge]
All arguments are optional
area_size is how far away it will shoot at you from, default is 800
update_rate is how often it checks for enemies, default is 1/2
timeout is how long it must wait between each fire before it can fire again, default is 1.5
ball_speed is how fast the ball moves, default is 1000
charge is the time between when it wants to shoot you and when it does (thus, charge time), default is 0

If you want to restrict access to the command, change the 'access' variable below to something like ACCESS_SLAY.
If you want to be able to move the pylons, change pylons_movable to true.

Changelog:
v1.0 (09/09/2006)
	*Initial version.
]]

assert( _file.Exists( "lua/ULib/init.lua" ), "UDodge needs ULib to run!" )
_OpenScript( "ULib/init.lua" )
assert( ULib.ULIB_VERSION >= 1, "UDodge requires ULib version 1 or higher to run!" )

local access = ACCESS_ALL
local pylons_movable = false

local offset = 50
local offset_up = vector3( 0, 0, 20 )

local pylons = {}
local maxplayers = _MaxPlayers()

local function register( callback, ... )
	local ents = _EntitiesFindByClass( "prop_combine_ball" )
	local num = table.getn( ents )
	if num < 1 then return end -- ERROR

	_EntFire( ents[ num ], "kill", "", 10 ) -- Give it a 10 second life span.
	_EntFire( ents[ num ], "addouput", "damage 50", 0 )
	ULib.print( unpack( arg ) )
	callback( ents[ num ], unpack( arg ) )
end

local function spawnBall( spawnpos, callback, ... )
	local entid = _EntCreate( "func_combine_ball_spawner" )
	if entid == 0 then ULib.print( "ERROR!" ) return end
	_EntSetModel( entid, "models/weapons/w_bugbait.mdl" )
	_EntSetKeyValue( entid, "maxspeed", "0" )
	_EntSetKeyValue( entid, "ballradius", "15" )
	_EntSetKeyValue( entid, "ballrespawntime", "0.01" )
	_EntSetKeyValue( entid, "ballcount", "1" )
	_EntSetPos( entid, spawnpos )
	_EntSpawn( entid )
	_EntFire( entid, "kill", "", 0.1 )

	AddTimer( 0.1, 1, register, callback, unpack( arg ) )
end

local function enemyFire( cball, target, speed, time )
	if time > 0 then -- We want to wait for charging sound to end.
		AddTimer( time, 1, enemyFire, cball, target, speed, 0 )
		return
	end

	local targetpos = _EntGetPos( target )
	targetpos = vecAdd( targetpos, offset_up )
	local cballpos = _EntGetPos( cball )
	local targetvel = _EntGetVelocity( target )

	local dist = vecSub( targetpos, cballpos )
	local timetotarget = vecLength( dist ) / speed
	local distcovered = vecMul( targetvel, timetotarget )
	local futurepos = vecAdd( targetpos, distcovered )
	local dir = vecNormalize( vecSub( futurepos, cballpos ) )

	ULib.play3DSound( "weapons/irifle/irifle_fire2.wav", cballpos )
	ULib.applyAccel( cball, speed, dir )
end


local function enemyCheck( pylonid )
	local pylon = pylons[ pylonid ]

	if pylon.timein > _CurTime() then return end -- Not past timeout

	local targets = {}
	local pos = _EntGetPos( pylon.entid )
	pos = vecAdd( pos, offset_up )
	local ents = _EntitiesFindInSphere( pos, pylon.area_size )
	for _, entid in ipairs( ents ) do
		if entid > 0 and entid <= maxplayers and _PlayerInfo( entid, "alive" ) then
			-- We're going to do a trace to make sure the weapon can see the player.
			local dir = vecSub( vecAdd( _EntGetPos( entid ), offset_up ), pos )
			_TraceLine( pos, dir, pylon.area_size, pylon.entid )
			if _TraceGetEnt() == entid then -- It can see them!
				table.insert( targets, entid )
			end
		end
	end

	if table.getn( targets ) > 0 then
		local entid = targets[ math.random( table.getn( targets ) ) ] -- Get a random target
		pylons[ pylonid ].timein = _CurTime() + pylon.timeout
		local off = vecNormalize( vecSub( _EntGetPos( entid ), pos ) )
		off = vecMul( off, offset )
		local spawnpos = vecAdd( pos, off )
		spawnpos = vecAdd( spawnpos, offset_up )
		ULib.play3DSound( "weapons/cguard/charging.wav", spawnpos )
		spawnBall( spawnpos, enemyFire, entid, pylon.speed, pylon.charge )
	end
	return
end

local function cc_makePylon( userid, args, argv, argc )
	if not ULib.mainUcl:query( userid, access ) then
		ULib.tsay( userid, "You do not have access to this command." )
		return
	end

	local area_size = tonumber( argv[ 1 ] ) or 800
	local update = tonumber( argv[ 2 ] ) or 1/2
	local timeout = tonumber( argv[ 3 ] ) or 1.5
	local speed = tonumber( argv[ 4 ] ) or 1000
	local charge = tonumber( argv[ 5 ] ) or 0

	PlayerLookTrace( userid, 4096 )
	local entid = ULib.makeProp( "models/props_c17/oildrum001.mdl", _TraceEndPos(), vector3( 1, 0, 0 ) )
	if not pylons_movable then
		_EntSetMoveType( entid, MOVETYPE_NONE )
	end

	local uprightid = _EntCreate( "phys_keepupright" )
	_EntSetKeyValue( uprightid, "attach1", ULib.generateEntName( entid ) )
	_EntSetKeyValue( uprightid, "angularlimit", "99.99" )
	_EntSpawn( uprightid )
	_EntActivate( uprightid )

	local pylonid = table.getn( pylons ) + 1
	local timerid = AddTimer( update, 0, enemyCheck, pylonid )
	table.insert( pylons, { entid=entid, area_size=area_size, timeout=timeout, speed=speed, timein=0, charge=charge } )

	ULib.callOnDelete( entid, function()
		HaltTimer( timerid )
		table.remove( pylons, pylonid )
	end )
end
ULib.CONCOMMAND( "makepylon", cc_makePylon )